/*----------------------------------------------------------------------
  File    : mlp.c
  Contents: multilayer perceptron management
  Author  : Christian Borgelt
  History : 2001.02.26 file created
            2001.03.10 first version completed
            2001.03.18 other weight update methods added
            2001.03.19 tabulated logistic function added
            2001.04.15 tangens hyperbolicus added as a #define option
            2001.04.16 output scaling added
            2001.05.20 extended functions added (attribute set/table)
            2001.07.17 adapted to modified module scan
            2001.09.11 some assertions added
            2002.06.04 input scaling/normalization added
            2002.06.18 sensitivity analysis functions added
            2002.06.19 bug in function mlp_result() fixed
            2002.09.12 scaling bug concerning numeric targets fixed
            2003.01.30 functionality of mlp_reg() extended
            2003.02.01 bug in mapping binary attributes fixed
            2003.02.07 reinitialization added to function mlp_range()
            2003.03.11 weight parameter added to function mlp_reg()
            2003.04.01 bug in function mlp_createx() fixed
            2003.08.06 another bug in mapping binary attributes fixed
            2003.10.24 bug in output range determination fixed
            2004.02.23 check of parabola orientation added to quick()
            2004.08.10 function mlp_deletex() added
            2004.08.11 adapted to new module attmap
            2004.08.12 adapted to new modules parse and nstats
            2007.01.10 usage of attribute maps changed
            2007.02.14 adapted to modified module attset
            2008.04.14 bug in function quick() fixed (parabola orient.)
            2013.06.12 bug in function mlp_desc() fixed (ranges)
            2013.08.28 bug in function mlp_parse() fixed (duplicate nst)
            2014.10.07 bug in function mlp_parse() fixed (missing nst)
            2018.05.11 AdaGrad, RMSProp, Adam, AdamCorr added
            2018.09.10 extended functions (attset, attmap) removed
            2018.09.13 bug in create() fixed (allocation size of pp)
----------------------------------------------------------------------*/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <limits.h>
#include <float.h>
#include <math.h>
#include <assert.h>
#include "mlp.h"
#ifdef STORAGE
#include "storage.h"
#endif

/*----------------------------------------------------------------------
  Preprocessor Definitions
----------------------------------------------------------------------*/
/* --- activation function and its derivative --- */
#ifdef MLP_TABFN                /* if to use a tabulated function */
#define TABSIZE     1024        /* size of interpolation table */
#define TABMAX      16.0        /* maximum (absolute) argument value */
#define TABSCALE    (TABSIZE/TABMAX)
#endif                          /* scale factor for table indices */

#if   defined(MLP_RELU)         /* actfn: leaky rectified maximum */
#define MACTFN(x)   ((x) > 0 ? (x) : 0.01*(x))
#define ACTFN(x)    MACTFN(x)   /* no tabulated version */
#define DERIV(x)    ((x) > 0 ? 1 : 0.01)
#define ACTMIN       0.0        /* minimal and maximal value */
#define ACTMAX       1.0        /* of the activation function */
#define ACTMID       0.5        /* middle value */
#define MLP_LRATE    0.05

#elif defined(MLP_TANH)         /* actfn: tangens hyperbolicus */
#define MACTFN(x)   (2.0/(1.0 +exp(-2*(x))) -1.0)
#ifdef MLP_TABFN                /* tabulated function */
#define ACTFN(x)    tanh(x)
#else                           /* direct computation */
#define ACTFN(x)    MACTFN(x)
#endif
/* #define DERIV(x)    ((1+(x))*(1-(x))) */
#define DERIV(x)    (1.0-(x)*(x))
#define ACTMIN      -1.0        /* minimal and maximal value */
#define ACTMAX       1.0        /* of the activation function */
#define ACTMID       0.0        /* middle value */
#define MLP_LRATE    0.05

#else                           /* default: logistic function */
#define MACTFN(x)   (1/(1 +exp(-(x))))
#ifdef MLP_TABFN                /* tabulated function */
#define ACTFN(x)    logistic(x)
#else                           /* direct computation */
#define ACTFN(x)    MACTFN(x)
#endif
#define DERIV(x)    ((x)*(1-(x)))
#define ACTMIN       0.0        /* minimal and maximal value */
#define ACTMAX       1.0        /* of the activation function */
#define ACTMID       0.5        /* middle value */
#define MLP_LRATE    0.2
#endif

#define NAMELEN     255         /* maximum target name length */

/* --- error codes --- */
#define E_ATTEXP    (-16)       /* attribute expected */
#define E_INPCNT    (-17)       /* invalid number of inputs */
#define E_OUTCNT    (-18)       /* invalid number of outputs */
#define E_LYRCNT    (-19)       /* invalid number of layers */

/*----------------------------------------------------------------------
  Type Definitions
----------------------------------------------------------------------*/
typedef void UPDATEFN (MLP* mlp); /* a weight update function */

/*----------------------------------------------------------------------
  Constants
----------------------------------------------------------------------*/
#ifdef MLP_PARSE
static const char *msgs[] = {   /* error messages */
  /*      0 to  -7 */  NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL,
  /*     -8 to -15 */  NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL,
  /* E_ATTEXP  -16 */  "#attribute expected instead of '%s'",
  /* E_INPCNT  -17 */  "#invalid number of inputs",
  /* E_OUTCNT  -18 */  "#invalid number of outputs",
  /* E_LYRCNT  -19 */  "#invalid number of layers",
};
#endif

/*----------------------------------------------------------------------
  Global Variables
----------------------------------------------------------------------*/
#ifdef MLP_TABFN
static double acttab[TABSIZE+1];/* tabulated activation function */
static int    init = 0;         /* whether table is initialized */
#endif

/*----------------------------------------------------------------------
  Activation Function
----------------------------------------------------------------------*/
#ifdef MLP_TABFN
#ifdef MLP_TANH

static double tanh (double x)
{                               /* --- table-based tanh function */
  int    i;                     /* table index */
  double y = fabs(x);           /* argument of tabulated function */

  if (y >= TABMAX)              /* if beyond the table range, */
    return (x > 0) ? 1 : -1;    /* return asymptotic values */
  i = (int)(y *= TABSCALE);     /* otherwise compute the table index */
  y = acttab[i] + (y-i) * (acttab[i+1] -acttab[i]);
  return (x < 0) ? -y : y;      /* interpolate between table values */
}  /* tanh() */

#else /*--------------------------------------------------------------*/

static double logistic (double x)
{                               /* --- table-based logistic function */
  int    i;                     /* table index */
  double y = fabs(x);           /* argument of tabulated function */

  if (y >= TABMAX)              /* if beyond the table range, */
    return (x > 0) ? 1 : 0;     /* return asymptotic values */
  i = (int)(y *= TABSCALE);     /* otherwise compute the table index */
  y = acttab[i] + (y-i) * (acttab[i+1] -acttab[i]);
  return (x < 0) ? 1-y : y;     /* interpolate between table values */
}  /* logistic() */

#endif  /* #ifdef MLP_TANH */
#endif  /* #ifdef MLP_TABFN */
/*----------------------------------------------------------------------
  Weight Update Functions
----------------------------------------------------------------------*/

static void standard (MLP *mlp)
{                               /* --- standard backpropagation */
  DIMID  i;                     /* loop variable */
  double *w, *c, *g;            /* to traverse the vectors */

  w = mlp->wgts;                /* get the necessary vectors and */
  g = mlp->grds;                /* traverse the connection weights */
  if      (mlp->moment > 0) {   /* if backpropagation with momentum */
    c = mlp->chgs;              /* get the vector of old changes */
    for (i = 0; i < mlp->wgtcnt; i++) {
      w[i] += c[i] = mlp->moment *c[i] - mlp->lrate *g[i];
      g[i]  = 0;                /* update the connection weights */
    } }                         /* and reinitialize the gradient */
  else if (mlp->moment < 0) {   /* if Nesterov's accelerated gradient */
    c = mlp->chgs;              /* get the vector of old gradients */
    for (i = 0; i < mlp->wgtcnt; i++) {
      c[i]  = -mlp->moment *c[i] +g[i];
      w[i] += mlp->lrate *(mlp->moment *c[i] -g[i]);
      g[i]  = 0;                /* update the connection weights */
    } }                         /* and reinitialize the gradient */
  else {                        /* if standard backpropagation */
    for (i = 0; i < mlp->wgtcnt; i++) {
      w[i] -= mlp->lrate *g[i]; /* update the connection weights */
      g[i]  = 0;                /* and reinitialize the gradient */
    }
  }
}  /* standard() */

/*--------------------------------------------------------------------*/

static void manhattan (MLP *mlp)
{                               /* --- Manhattan training */
  DIMID  i;                     /* loop variable */
  double *w, *g;                /* to traverse the vectors */

  w = mlp->wgts;                /* get the necessary vectors and */
  g = mlp->grds;                /* traverse the connection weights */
  for (i = 0; i < mlp->wgtcnt; i++) {
    if      (g[i] > 0) w[i] -= mlp->lrate;
    else if (g[i] < 0) w[i] += mlp->lrate;
    g[i] = 0;                   /* update the connection weights */
  }                             /* and reinitialize the gradient */
}  /* manhattan() */

/*--------------------------------------------------------------------*/

static void adaptive (MLP *mlp)
{                               /* --- super self-adaptive backprop. */
  DIMID  i;                     /* loop variable */
  double *w, *c, *g, *p;        /* to traverse the vectors */
  double t;                     /* temporary buffer */

  w = mlp->wgts; c = mlp->chgs; /* get the necessary vectors and */
  g = mlp->grds; p = mlp->bf1s; /* traverse the connection weights */
  for (i = 0; i < mlp->wgtcnt; i++) {
    if      (g[i] > 0) t =  p[i];
    else if (g[i] < 0) t = -p[i];
    else               t =  0;  /* check directions of the changes */
    if      (t > 0) {           /* if gradients have the same sign, */
      c[i] *= mlp->growth;      /* increase the step width */
      if (c[i] > mlp->maxchg) c[i] = mlp->maxchg;
      p[i] = g[i]; }            /* note the current gradient */
    else if (t < 0) {           /* if gradients have opposite signs, */
      c[i] *= mlp->shrink;      /* decrease the step width */
      if (c[i] < mlp->minchg) c[i] = mlp->minchg;
      p[i] = 0; }               /* suppress a change in the next step */
    else {                      /* if one gradient is zero, */
      p[i] = g[i]; }            /* only note the current gradient */
    w[i] -= c[i] *g[i];         /* update the connection weight */
    g[i] = 0;                   /* and reinitialize the gradient */
  }
}  /* adaptive() */

/*--------------------------------------------------------------------*/

static void resilient (MLP *mlp)
{                               /* --- resilient backpropagation */
  DIMID  i;                     /* loop variable */
  double *w, *c, *g, *p;        /* to traverse the vectors */
  double t;                     /* temporary buffer */

  w = mlp->wgts; c = mlp->chgs; /* get the necessary vectors and */
  g = mlp->grds; p = mlp->bf1s; /* traverse the connection weights */
  for (i = 0; i < mlp->wgtcnt; i++) {
    if      (g[i] > 0) t =  p[i];
    else if (g[i] < 0) t = -p[i];
    else               t =  0;  /* check directions of the changes */
    if      (t > 0) {           /* if gradients have the same sign, */
      c[i] *= mlp->growth;      /* increase the update value */
      if (c[i] > mlp->maxchg) c[i] = mlp->maxchg;
      p[i] = g[i]; }            /* note the current gradient */
    else if (t < 0) {           /* if gradients have opposite signs, */
      c[i] *= mlp->shrink;      /* decrease the update value */
      if (c[i] < mlp->minchg) c[i] = mlp->minchg;
      p[i] = 0; }               /* suppress a change in the next step */
    else {                      /* if one gradient is zero, */
      p[i] = g[i]; }            /* only note the current gradient */
    if      (g[i] > 0) w[i] -= c[i];  /* update the weight */
    else if (g[i] < 0) w[i] += c[i];  /* (but use only the sign) */
    g[i] = 0;                   /* reinitialize the gradient */
  }
}  /* resilient() */

/*--------------------------------------------------------------------*/

static void quick (MLP *mlp)
{                               /* --- quick backpropagation */
  DIMID  i;                     /* loop variable */
  double *w, *c, *g, *p;        /* to traverse the vectors */
  double m;                     /* maximal fraction of new derivative */
  double t;                     /* temporary buffer */

  m = mlp->growth /(mlp->growth +1);
  w = mlp->wgts; c = mlp->chgs; /* get the necessary vectors and */
  g = mlp->grds; p = mlp->bf1s; /* traverse the connection weights */
  for (i = 0; i < mlp->wgtcnt; i++) {
    t = p[i] -g[i];             /* compute the change of the gradient */
    if      (t * c[i] >= 0)     /* if the parabola opens downwards, */
      c[i] = -mlp->lrate *g[i]; /* do a normal backpropagation step */
    else if (p[i] > 0) {        /* if previous gradient was positive */
      if (g[i] < m *p[i])       /* compute the factor for a jump */
        c[i] *= g[i] /t;        /* to the minimum (apex of parabola) */
      else                      /* if the growth factor would become */
        c[i] *= mlp->growth;    /* too large, use the maximal factor */
      if (g[i] > 0)             /* if steps are in the same dir., */
        c[i] -= mlp->lrate *g[i]; } /* add a normal backprop. step */
    else if (p[i] < 0) {        /* if previous gradient was negative */
      if (g[i] > m *p[i])       /* compute the factor for a jump */
        c[i] *= g[i] /t;        /* to the minimum (apex of parabola) */
      else                      /* if the growth factor would become */
        c[i] *= mlp->growth;    /* too large, use the maximal factor */
      if (g[i] < 0)             /* if steps are in the same dir., */
        c[i] -= mlp->lrate *g[i]; } /* add a normal backprop. step */
    else                        /* if this is the first update, */
      c[i] = -mlp->lrate *g[i]; /* do a normal backpropagation step */
    if      (c[i] >  mlp->maxchg) c[i] =  mlp->maxchg;  /* clamp the */
    else if (c[i] < -mlp->maxchg) c[i] = -mlp->maxchg;  /* change */
    w[i] += c[i];               /* adapt the connection weight, */
    p[i] = g[i]; g[i] = 0;      /* note the current gradient */
  }                             /* and reinitialize it */
}  /* quick() */

/*--------------------------------------------------------------------*/

static void adagrad (MLP *mlp)
{                               /* --- adaptive subgradient descent */
  DIMID  i;                     /* loop variable */
  double *w, *c, *g, *v;        /* to traverse the vectors */
  double t;                     /* buffer for current change */

  w = mlp->wgts;                /* get the necessary vectors and */
  g = mlp->grds;                /* traverse the connection weights, */
  v = mlp->bf1s;                /* gradients and squared gradients */
  if      (mlp->moment > 0) {   /* if backpropagation with momentum */
    c = mlp->chgs;              /* get the vector of old changes */
    for (i = 0; i < mlp->wgtcnt; i++) {
      v[i] += g[i]*g[i];        /* sum the squared gradients */
      t     = mlp->lrate *(g[i] /(sqrt(v[i]) +mlp->epsilon));
      w[i] += c[i] = mlp->moment *c[i] - t;
      g[i]  = 0;                /* update the connection weights */
    } }                         /* and reinitialize the gradient */
  else if (mlp->moment < 0) {   /* if Nesterov's accelerated gradient */
    c = mlp->chgs;              /* get the vector of old changes */
    for (i = 0; i < mlp->wgtcnt; i++) {
      v[i] += g[i]*g[i];        /* sum the squared gradients */
      t     = g[i] /(sqrt(v[i]) +mlp->epsilon);
      c[i]  = -mlp->moment *c[i] +t;
      w[i] += mlp->lrate *(mlp->moment *c[i] -t);
      g[i]  = 0;                /* update the connection weights */
    } }                         /* and reinitialize the gradient */
  else {                        /* if standard backpropagation */
    for (i = 0; i < mlp->wgtcnt; i++) {
      v[i] += g[i]*g[i];        /* sum the squared gradients */
      w[i] -= mlp->lrate *(g[i] /(sqrt(v[i]) +mlp->epsilon));
      g[i]  = 0;                /* update the connection weights */
    }                           /* and reinitialize the gradient */
  }
}  /* adagrad() */

/*--------------------------------------------------------------------*/

static void rmsprop (MLP *mlp)
{                               /* --- root mean squared gradients */
  DIMID  i;                     /* loop variable */
  double *w, *c, *g, *v;        /* to traverse the vectors */
  double t;                     /* buffer for current change */

  w = mlp->wgts;                /* get the necessary vectors and */
  g = mlp->grds;                /* traverse the connection weights, */
  v = mlp->bf1s;                /* gradients and squared gradients */
  if      (mlp->moment > 0) {   /* if backpropagation with momentum */
    c = mlp->chgs;              /* get the vector of old changes */
    for (i = 0; i < mlp->wgtcnt; i++) {
      v[i]  = mlp->dvar *v[i] +(1 -mlp->dvar) *(g[i]*g[i]);
      t     = mlp->lrate *(g[i] /(sqrt(v[i]) +mlp->epsilon));
      w[i] += c[i] = mlp->moment *c[i] - t;
      g[i]  = 0;                /* update the connection weights */
    } }                         /* and reinitialize the gradient */
  else if (mlp->moment < 0) {   /* if Nesterov's accelerated gradient */
    c = mlp->chgs;              /* get the vector of old changes */
    for (i = 0; i < mlp->wgtcnt; i++) {
      v[i]  = mlp->dvar *v[i] +(1 -mlp->dvar) *(g[i]*g[i]);
      t     = g[i] /(sqrt(v[i]) +mlp->epsilon);
      c[i]  = -mlp->moment *c[i] +t;
      w[i] += mlp->lrate *(mlp->moment *c[i] -t);
      g[i]  = 0;                /* update the connection weights */
    } }                         /* and reinitialize the gradient */
  else {                        /* if standard backpropagation */
    for (i = 0; i < mlp->wgtcnt; i++) {
      v[i]  = mlp->dvar *v[i] +(1 -mlp->dvar) *(g[i]*g[i]);
      w[i] -= mlp->lrate *(g[i] /(sqrt(v[i]) +mlp->epsilon));
      g[i]  = 0;                /* update the connection weights */
    }                           /* and reinitialize the gradient */
  }
}  /* rmsprop() */

/*--------------------------------------------------------------------*/

static void adadelta (MLP *mlp)
{                               /* --- rmsprop with 2nd order term */
  DIMID  i;                     /* loop variable */
  double *w, *c, *g, *v;        /* to traverse the vectors */
  double t;                     /* buffer */

  w = mlp->wgts;                /* get the necessary vectors and */
  g = mlp->grds;                /* traverse the connection weights, */
  v = mlp->bf1s;                /* gradients and squared gradients */
  c = mlp->chgs;                /* and the vector of old changes */
  for (i = 0; i < mlp->wgtcnt; i++) {
    v[i]  = mlp->dvar *v[i] +(1 -mlp->dvar) *(g[i]*g[i]);
    w[i] -= t = ((sqrt(c[i] +mlp->epsilon))
                /(sqrt(v[i] +mlp->epsilon))) *g[i];
    c[i]  = mlp->dvar *c[i] +(1 -mlp->dvar) *t*t;
    g[i]  = 0;                /* update the connection weights */
  }
}  /* adadelta() */

/*--------------------------------------------------------------------*/

static void adam (MLP *mlp)
{                               /* --- adaptive moment estimation */
  DIMID  i;                     /* loop variable */
  double *w, *g, *m, *v;        /* to traverse the vectors */

  w = mlp->wgts;                /* get the necessary vectors and */
  g = mlp->grds;                /* traverse the connection weights, */
  m = mlp->chgs;                /* gradients and their mean estimate */
  v = mlp->bf1s;                /* as well as squared gradients */
  for (i = 0; i < mlp->wgtcnt; i++) {
    m[i]  = mlp->dmean *m[i] +(1 -mlp->dmean) * g[i];
    v[i]  = mlp->dvar  *v[i] +(1 -mlp->dvar)  *(g[i]*g[i]);
    w[i] -= mlp->lrate *(m[i] /(sqrt(v[i]) +mlp->epsilon));
    g[i]  = 0;                  /* update the connection weights */
  }                             /* and reinitialize the gradient */
}  /* adam() */

/*--------------------------------------------------------------------*/

static void adambc (MLP *mlp)
{                               /* --- adam with bias correction */
  DIMID  i;                     /* loop variable */
  double *w, *g, *m, *v;        /* to traverse the vectors */
  double cm, cv;                /* mean/variance correction factors */

  w = mlp->wgts;                /* get the necessary vectors and */
  g = mlp->grds;                /* traverse the connection weights, */
  m = mlp->chgs;                /* gradients and their mean estimate */
  v = mlp->bf1s;                /* as well as squared gradients */
  cm = 1.0/(1.0 -pow(mlp->dmean, mlp->step));
  cv = 1.0/(1.0 -pow(mlp->dvar,  mlp->step));
  for (i = 0; i < mlp->wgtcnt; i++) {
    m[i]  = mlp->dmean *m[i] +(1 -mlp->dmean) * g[i];
    v[i]  = mlp->dvar  *v[i] +(1 -mlp->dvar)  *(g[i]*g[i]);
    w[i] -= mlp->lrate *(m[i]*cm /(sqrt(v[i]*cv) +mlp->epsilon));
    g[i]  = 0;                  /* update the connection weights */
  }                             /* and reinitialize the gradient */
}  /* adambc() */

/*--------------------------------------------------------------------*/

static void nadam (MLP *mlp)
{                               /* --- Nesterov-accelerated adam */
  DIMID  i;                     /* loop variable */
  double *w, *g, *m, *v;        /* to traverse the vectors */

  w = mlp->wgts;                /* get the necessary vectors and */
  g = mlp->grds;                /* traverse the connection weights, */
  m = mlp->chgs;                /* gradients and their mean estimate */
  v = mlp->bf1s;                /* as well as squared gradients */
  for (i = 0; i < mlp->wgtcnt; i++) {
    m[i]  = mlp->dmean *m[i] +(1 -mlp->dmean) * g[i];
    v[i]  = mlp->dvar  *v[i] +(1 -mlp->dvar)  *(g[i]*g[i]);
    w[i] -= mlp->lrate *((mlp->dmean*m[i] +(1 -mlp->dmean)*g[i])
                        /(sqrt(v[i]) +mlp->epsilon));
    g[i]  = 0;                  /* update the connection weights */
  }                             /* and reinitialize the gradient */
}  /* nadam() */

/*--------------------------------------------------------------------*/

static void nadambc (MLP *mlp)
{                               /* --- nadam with bias correction */
  DIMID  i;                     /* loop variable */
  double *w, *g, *m, *v;        /* to traverse the vectors */
  double cg, cm, cv;            /* mean/variance correction factors */

  w = mlp->wgts;                /* get the necessary vectors and */
  g = mlp->grds;                /* traverse the connection weights, */
  m = mlp->chgs;                /* gradients and their mean estimate */
  v = mlp->bf1s;                /* as well as squared gradients */
  cv = pow(mlp->dmean, mlp->step);
  cg = (1-mlp->dmean)/(1.0 -cv);
  cm = mlp->dmean/(1.0 -cv*mlp->dmean);
  cv = 1.0/(1.0 -pow(mlp->dvar, mlp->step));
  for (i = 0; i < mlp->wgtcnt; i++) {
    m[i]  = mlp->dmean *m[i] +(1 -mlp->dmean) * g[i];
    v[i]  = mlp->dvar  *v[i] +(1 -mlp->dvar)  *(g[i]*g[i]);
    w[i] -= mlp->lrate *((cm*m[i] +cg*g[i])
                        /(sqrt(v[i]*cv) +mlp->epsilon));
    g[i]  = 0;                  /* update the connection weights */
  }                             /* and reinitialize the gradient */
}  /* nadambc() */

/*--------------------------------------------------------------------*/

static void adamax (MLP *mlp)
{                               /* --- adam with l_inf estimation */
  DIMID  i;                     /* loop variable */
  double *w, *g, *m, *v;        /* to traverse the vectors */
  double t, u;                  /* buffers */

  w = mlp->wgts;                /* get the necessary vectors and */
  g = mlp->grds;                /* traverse the connection weights, */
  m = mlp->chgs;                /* gradients and their mean estimate */
  v = mlp->bf1s;                /* as well as squared gradients */
  for (i = 0; i < mlp->wgtcnt; i++) {
    m[i]  = mlp->dmean *m[i] +(1 -mlp->dmean) *g[i];
    t = fabs(g[i]); u = mlp->dvar *v[i];
    v[i]  = (t > u) ? t : u;    /* update mean with maximum scheme */
    w[i] -= mlp->lrate *(m[i] /(v[i] +mlp->epsilon));
    g[i]  = 0;                  /* update the connection weights */
  }                             /* and reinitialize the gradient */
}  /* adamax() */

/*--------------------------------------------------------------------*/

static void adamaxbc (MLP *mlp)
{                               /* --- adam with bias corr. and l_inf */
  DIMID  i;                     /* loop variable */
  double *w, *g, *m, *v;        /* to traverse the vectors */
  double cm;                    /* mean correction factor */
  double t, u;                  /* buffers */

  w = mlp->wgts;                /* get the necessary vectors and */
  g = mlp->grds;                /* traverse the connection weights, */
  m = mlp->chgs;                /* gradients and their mean estimate */
  v = mlp->bf1s;                /* as well as squared gradients */
  cm = 1.0/(1.0 -pow(mlp->dmean, mlp->step));
  for (i = 0; i < mlp->wgtcnt; i++) {
    m[i]  = mlp->dmean *m[i] +(1 -mlp->dmean) *g[i];
    t = fabs(g[i]); u = mlp->dvar *v[i];
    v[i]  = (t > u) ? t : u;    /* update mean with maximum scheme */
    w[i] -= mlp->lrate *(m[i]*cm /(v[i] +mlp->epsilon));
    g[i]  = 0;                  /* update the connection weights */
  }                             /* and reinitialize the gradient */
}  /* adamaxbc() */

/*--------------------------------------------------------------------*/

static void amsgrad (MLP *mlp)
{                               /* --- adam with variance maximum */
  DIMID  i;                     /* loop variable */
  double *w, *g, *m, *v, *y;    /* to traverse the vectors */

  w = mlp->wgts;                /* get the necessary vectors and */
  g = mlp->grds;                /* traverse the connection weights, */
  m = mlp->chgs;                /* gradients and their mean estimate */
  v = mlp->bf1s;                /* as well as squared gradients */
  y = mlp->bf2s;                /* and their maximum */
  for (i = 0; i < mlp->wgtcnt; i++) {
    m[i]  = mlp->dmean *m[i] +(1 -mlp->dmean) *g[i];
    v[i]  = mlp->dvar  *v[i] +(1 -mlp->dvar) *(g[i]*g[i]);
    if (v[i] > y[i]) y[i] = v[i];
    w[i] -= mlp->lrate *(m[i] /(sqrt(y[i]) +mlp->epsilon));
    g[i]  = 0;                  /* update the connection weights */
  }                             /* and reinitialize the gradient */
}  /* amsgrad() */

/*--------------------------------------------------------------------*/

static UPDATEFN *updatefn[] = {
  /* MLP_STANDARD   0 */  standard,
  /* MLP_MANHATTAN  1 */  manhattan,
  /* MLP_ADAPTIVE   2 */  adaptive,
  /* MLP_RESILIENT  3 */  resilient,
  /* MLP_QUICK      4 */  quick,
  /* MLP_ADAGRAD    5 */  adagrad,
  /* MLP_RMSPROP    6 */  rmsprop,
  /* MLP_ADADELTA   7 */  adadelta,
  /* MLP_ADAM       8 */  adam,
  /* MLP_ADAMBC     9 */  adambc,
  /* MLP_NADAM     10 */  nadam,
  /* MLP_NADAMBC   11 */  nadambc,
  /* MLP_ADAMAX    12 */  adamax,
  /* MLP_ADAMAXBC  13 */  adamaxbc,
  /* MLP_AMSGRAD   14 */  amsgrad,
};                              /* list of weight update functions */

/*----------------------------------------------------------------------
  Main Functions
----------------------------------------------------------------------*/

static MLP* create (int lyrcnt, DIMID *ucnts)
{                               /* --- create a multilayer perceptron */
  int    l;                     /* loop variable for layers */
  DIMID  k, n;                  /* loop variable for weights, buffer */
  MLP    *mlp;                  /* created multilayer perceptron */
  double **pp, *p, *o;          /* to traverse the allocated vectors */

  #ifdef MLP_TABFN              /* if to use tabulated logistic fn. */
  if (!_init) {                 /* if table has not been initialized */
    for (k = TABSIZE; --k >= 0; )
      acttab[k] = MACTFN(k *(1/TABSCALE));
    acttab[TABSIZE] = 1;        /* precompute a table for */
    init = 1;                   /* the activation function and */
  }                             /* set the initialization flag */
  #endif  /* #ifdef MLP_TABFN */

  assert((lyrcnt >= 2)          /* check the function arguments */
  &&     (lyrcnt <= MLP_MAXLAYER) && ucnts);
  mlp = (MLP*)malloc(sizeof(MLP) +(size_t)(lyrcnt-2) *sizeof(MLPLAYER));
  if (!mlp) return NULL;        /* allocate the base structure */
  mlp->lyrcnt = lyrcnt;         /* note the number of layers and */
  mlp->inpcnt = ucnts[0];       /* the number of inputs/outputs */
  mlp->outcnt = ucnts[lyrcnt-1];
  assert(mlp->outcnt > 0);      /* check the number of outputs */
  for (k = 0, l = lyrcnt; --l > 0; )
    k += ucnts[l];              /* determine the number of units */
  mlp->unitcnt = k +ucnts[0];   /* and allocate matrix vectors */
  pp = (double**)malloc(5 *(size_t)k *sizeof(double*));
  if (!pp) { free(mlp); return NULL; }
  for (n = 0, l = lyrcnt; --l > 0; )
    n += ucnts[l] *(ucnts[l-1] +1); /* det. the number of weights */
  mlp->wgtcnt = n;                  /* and num. of inputs/outputs */
  p = (double*)malloc((2*(size_t)k +5*(size_t)n
                      +1*(size_t)mlp->inpcnt
                      +7*(size_t)mlp->outcnt) *sizeof(double));
  if (!p) { free(pp); free(mlp); return NULL; }
  mlp->mins = p; p += mlp->outcnt;     /* set the vectors for the */
  mlp->maxs = p; p += mlp->outcnt;     /* minimal/maximal outputs, */
  mlp->offs = p; p += mlp->outcnt;     /* the offsets, */
  mlp->scls = p; p += mlp->outcnt;     /* the scaling factors */
  mlp->recs = p; p += mlp->outcnt;     /* and their reciprocals, */
  mlp->scos = p; p += mlp->outcnt;     /* the scaled outputs, */
  mlp->trgs = p; p += mlp->outcnt;     /* the target outputs, */
  mlp->inps = o = p; p += mlp->inpcnt; /* and the input vector */
  lyrcnt -= 1;                  /* traverse the network layers */
  for (l = 0; l < lyrcnt; l++){ /* and init. inputs and outputs */
    mlp->layers[l].inpcnt = ucnts[l];
    mlp->layers[l].outcnt = k = ucnts[l+1];
    mlp->layers[l].inps   = o; o = p;
    mlp->layers[l].outs   = p; p += k;
    mlp->layers[l].errs   = p; p += k;
  }                             /* set the layer specific vectors */
  mlp->outs = mlp->layers[l-1].outs; /* set output and error vector */
  mlp->errs = mlp->layers[l-1].errs; /* of the network as a whole */
  mlp->wgts = p;                /* note the weight vector */
  for (l = 0; l < lyrcnt; l++){ /* traverse the layers */
    mlp->layers[l].wgts = pp; n = ucnts[l] +1;
    for (k = ucnts[l+1]; --k >= 0; ) { *pp++ = p; p += n; }
  }                             /* set the weight matrix lines */
  mlp->grds = p;                /* note the gradient vector */
  for (l = 0; l < lyrcnt; l++){ /* traverse the layers */
    mlp->layers[l].grds = pp; n = ucnts[l] +1;
    for (k = ucnts[l+1]; --k >= 0; ) { *pp++ = p; p += n; }
  }                             /* set the gradient matrix lines */
  mlp->chgs = p;                /* note the weight change vector */
  for (l = 0; l < lyrcnt; l++){ /* traverse the layers */
    mlp->layers[l].chgs = pp; n = ucnts[l] +1;
    for (k = ucnts[l+1]; --k >= 0; ) { *pp++ = p; p += n; }
  }                             /* set the weight change matrix lines */
  mlp->bf1s = p;                /* note the buffer vectors */
  for (l = 0; l < lyrcnt; l++){ /* traverse the layers */
    mlp->layers[l].bf1s = pp; n = ucnts[l] +1;
    for (k = ucnts[l+1]; --k >= 0; ) { *pp++ = p; p += n; }
  }                             /* set the buffer vectors */
  mlp->bf2s = p;                /* note the buffer vectors */
  for (l = 0; l < lyrcnt; l++){ /* traverse the layers */
    mlp->layers[l].bf2s = pp; n = ucnts[l] +1;
    for (k = ucnts[l+1]; --k >= 0; ) { *pp++ = p; p += n; }
  }                             /* set the buffer vectors */
  for (k = 0; k < mlp->outcnt; k++) {
    mlp->mins[k] =  INFINITY;   /* traverse the outputs and */
    mlp->maxs[k] = -INFINITY;   /* initialize the output ranges */
  }
  mlp->nst     = NULL;          /* clear the norm. statistics */
  mlp->method  = MLP_STANDARD;  /* set default update method */
  mlp->raise   = 0.0;           /* and set default values */
  mlp->lrate   = MLP_LRATE;     /* for the learning rate */
  mlp->moment  = 0.0;           /* and all other parameters */
  mlp->growth  = 1.2;
  mlp->shrink  = 0.5;
  mlp->minchg  = 1e-6;
  mlp->maxchg  = 16.0;
  mlp->dmean   = 0.9;
  mlp->dvar    = 0.999;
  mlp->epsilon = 1e-8;
  mlp->step    = 0.0;
  mlp->decay   = 1.0;
  return mlp;                   /* return the created network */
}  /* create() */

/*--------------------------------------------------------------------*/

MLP* mlp_create (int lyrcnt, DIMID *ucnts)
{                               /* --- create a multilayer perceptron */
  MLP *mlp;                     /* created multilayer perceptron */

  mlp = create(lyrcnt, ucnts);  /* create a multilayer perceptron */
  if (!mlp) return NULL;        /* and the normalization statistics */
  mlp->nst = nst_create(ucnts[0]);
  if (!mlp->nst) { mlp_delete(mlp); return NULL; }
  return mlp;                   /* return the created network */
}  /* mlp_create() */

/*--------------------------------------------------------------------*/

void mlp_delete (MLP *mlp)
{                               /* --- delete a multilayer perceptron */
  assert(mlp);                  /* check the function arguments */
  free(mlp->mins);              /* delete the weight vectors etc., */
  free(mlp->layers[0].wgts);    /* the weight matrix vectors and */
  nst_delete(mlp->nst);         /* the normalization statistics */
  free(mlp);                    /* delete the base structure */
}  /* mlp_delete() */

/*--------------------------------------------------------------------*/

void mlp_range (MLP *mlp, DIMID unit, double min, double max)
{                               /* --- set range of output values */
  assert(mlp && (unit < mlp->outcnt));     /* check the arguments */
  if (unit < 0) {               /* if to reinitialize the ranges */
    for (unit = mlp->outcnt; --unit >= 0; ) {
      mlp->mins[unit] =  INFINITY; /* traverse all output units and */
      mlp->maxs[unit] = -INFINITY; /* set the maximal range w.r.t. */
    } }                         /* the used activation function */
  else {                        /* if to set a specific range */
    mlp->mins[unit] = min;      /* set minimal */
    mlp->maxs[unit] = max;      /* and maximal value */
  }                             /* for the specified output */
}  /* mlp_range() */

/*--------------------------------------------------------------------*/

void mlp_expand (MLP *mlp, DIMID unit, double factor)
{                               /* --- expand range of output values */
  double t;                     /* change of minimal/maximal value */

  assert(mlp                    /* check the function arguments */
  &&    (unit >= 0) && (unit < mlp->outcnt) && (factor > 0));
  t = (mlp->maxs[unit] -mlp->mins[unit]) *(factor -1) *0.5;
  mlp->mins[unit] -= t;         /* adapt minimal and */
  mlp->maxs[unit] += t;         /* maximal output value */
}  /* mlp_expand() */

/*--------------------------------------------------------------------*/

void mlp_reg (MLP *mlp, const double *inps,
                        const double *trgs, double weight)
{                               /* --- register input/target pairs */
  DIMID i;                      /* loop variable */

  assert(mlp);                  /* check the function arguments */
  nst_reg(mlp->nst, inps, weight);
  if (!trgs) return;            /* check if target values are given */
  for (i = 0; i < mlp->outcnt; i++) {
    if (trgs[i] < mlp->mins[i]) mlp->mins[i] = trgs[i];
    if (trgs[i] > mlp->maxs[i]) mlp->maxs[i] = trgs[i];
  }                             /* adapt the ranges of output values */
}  /* mlp_reg() */

/*--------------------------------------------------------------------*/

void mlp_init (MLP *mlp, double rand(void), double range)
{                               /* --- init. multilayer perceptron */
  int      l;                   /* loop variable  for layers */
  DIMID    k, n;                /* loop variables for weights */
  MLPLAYER *layer;              /* to traverse the network layers */
  double   *wgt;                /* to traverse the weight vectors */

  assert(mlp && rand && (range > 0));   /* check the function args. */
  range *= 2;                   /* compute the full range of values */
  layer = mlp->layers;          /* traverse the network layers */
  for (l = mlp->lyrcnt-1; --l >= 0; ++layer) {
    for (k = layer->outcnt; --k >= 0; ) {
      wgt = layer->wgts[k];     /* traverse the units of the layer */
      wgt[n = layer->inpcnt] = 0;  /* set bias value to zero */
      while (--n >= 0) wgt[n] = range *(rand()-0.5);
    }                           /* initialize the weights uniformly */
  }                             /* from the interval [-e,e] */
}  /* mlp_init() */

/*--------------------------------------------------------------------*/

void mlp_jog (MLP *mlp, double rand(void), double range)
{                               /* --- jog weights of m.l. perceptron */
  DIMID i;                      /* loop variable */

  assert(mlp && rand && (range > 0));   /* check the function args. */
  range *= 2;                   /* compute the full range of values */
  for (i = 0; i < mlp->wgtcnt; i++)      /* jog the weights */
    mlp->wgts[i] += range *(rand()-0.5); /* with random values */
}  /* mlp_jog() */

/*--------------------------------------------------------------------*/

void mlp_setup (MLP *mlp)
{                               /* --- set up multilayer perceptron */
  DIMID  i;                     /* loop variable */
  double t;                     /* temporary buffer */
  double *m, *n, *o, *r, *s;    /* to access the arrays */

  assert(mlp);                  /* check the function argument */
  memset(mlp->grds, 0, (size_t)mlp->wgtcnt *sizeof(double));
  if ((mlp->method == MLP_RESILIENT)
  ||  (mlp->method == MLP_ADAPTIVE)) {
    for (i = 0; i < mlp->wgtcnt; i++)
      mlp->chgs[i] = mlp->lrate; }    /* initialize weight changes */
  else                          /* and clear gradients and buffers */
    memset(mlp->chgs, 0, (size_t)mlp->wgtcnt *sizeof(double));
  memset(mlp->bf1s, 0, (size_t)mlp->wgtcnt *sizeof(double));
  memset(mlp->bf2s, 0, (size_t)mlp->wgtcnt *sizeof(double));
  o = mlp->offs;                /* get the vectors of offsets */
  s = mlp->scls; r = mlp->recs; /* and (inverse) scaling factors */
  m = mlp->mins; n = mlp->maxs; /* as well as minimum and maximum */
  for (i = 0; i < mlp->outcnt; i++) {
    if (m[i] > n[i]) { m[i] = ACTMIN; n[i] = ACTMAX; }
    t = n[i] -m[i];             /* compute the range and */
    s[i] = t /(ACTMAX -ACTMIN); /* the scaling factor */
    t = (t > 0) ? 1/t : 1.0;    /* avoid infinity */
    r[i] = t *(ACTMAX -ACTMIN); /* comput reciprocal scaling factor */
    o[i] = m[i] -ACTMIN *s[i];  /* and the offset */
  }                             /* (linear transformation parameters) */
  mlp->step = 0;                /* init. the training step */
}  /* mlp_setup() */

/*--------------------------------------------------------------------*/

void mlp_exec (MLP *mlp, const double *inps, double *outs)
{                               /* --- execute multilayer perceptron */
  int      l;                   /* loop variable  for layers */
  DIMID    k, n;                /* loop variables for weights */
  MLPLAYER *layer;              /* to traverse the network layers */
  double   *wgt;                /* to traverse the weight vectors */
  double   net;                 /* sum of weighted inputs */

  assert(mlp);                  /* check the function arguments */
  if (inps)                     /* normalize the input vector */
    nst_norm(mlp->nst, inps, mlp->inps);
  layer = mlp->layers;          /* traverse the network layers */
  for (l = mlp->lyrcnt-1; --l >= 0; ++layer) {
    for (k = layer->outcnt; --k >= 0; ) {
      wgt = layer->wgts[k];     /* traverse the units of the layer */
      net = wgt[n = layer->inpcnt];
      while (--n >= 0) net += layer->inps[n] *wgt[n];
      layer->outs[k] = ACTFN(net);
    }                           /* sum the weighted inputs and */
  }                             /* compute the activation (output) */
  for (k = 0; k < mlp->outcnt; k++) /* apply output transformation */
    mlp->scos[k] = mlp->outs[k] *mlp->scls[k] +mlp->offs[k];
  if (outs)                     /* copy outputs to result vector */
    memcpy(outs, mlp->scos, (size_t)mlp->outcnt *sizeof(double));
}  /* mlp_exec() */

/*--------------------------------------------------------------------*/

double mlp_error (MLP *mlp, const double *trgs)
{                               /* --- compute sum of squared errors */
  DIMID  k;                     /* loop variable */
  double sse = 0, d;            /* sum of squared errors, buffer */

  assert(mlp);                  /* check the function arguments */
  if (!trgs) trgs = mlp->trgs;  /* if not targets, use intern. vector */
  for (k = 0; k < mlp->outcnt; k++) {
    mlp->errs[k] = mlp->recs[k] *(d = trgs[k] -mlp->scos[k]);
    sse += d*d;                 /* compute the output errors */
  }                             /* (target - scaled computed output) */
  return sse;                   /* return sum of squared errors */
}  /* mlp_error() */

/*--------------------------------------------------------------------*/

double mlp_bkprop (MLP *mlp, const double *trgs)
{                               /* --- backpropagate errors */
  int      l;                   /* loop variable  for layers */
  DIMID    k, n;                /* loop variables for weights */
  MLPLAYER *layer;              /* to traverse the network layers */
  double   *w, *g;              /* to traverse the weights/gradients */
  double   *e;                  /* to traverse the errors */
  double   delta;               /* temporary buffer */
  double   sse;                 /* sum of squared errors */
  double   raise = mlp->raise;  /* raise value for derivative */

  assert(mlp);                  /* check the function arguments */
  sse = mlp_error(mlp, trgs);   /* compute sum of squared errors */
  for (l = mlp->lyrcnt-2; --l >= 0; ) {
    memset(mlp->layers[l].errs, 0,
          (size_t)mlp->layers[l].outcnt *sizeof(double));
  }                             /* clear the backpropagated errors */
  for (l = mlp->lyrcnt-1; --l >  0; ) {
    layer = mlp->layers +l;     /* traverse the units of each layer */
    for (k = 0; k < layer->outcnt; k++) {
      delta = layer->errs[k] * (DERIV(layer->outs[k]) +raise);
      g     = layer->grds[k];   /* first process the offset gradient */
      g[n = layer->inpcnt] -= delta;
      while (--n >= 0)          /* then process the weight gradients */
        g[n] -= layer->inps[n] * delta;
      e = (layer-1)->errs;      /* traverse the errors */
      w = layer->wgts[k];       /* and the connection weights */
      for (n = 0; n < layer->inpcnt; n++)
        e[n] += w[n] * delta;   /* propagate the errors back */
    }                           /* to the preceding layer */
  }
  layer = mlp->layers;          /* traverse the first hidden layer */
  for (k = 0; k < layer->outcnt; k++) {
    delta = layer->errs[k] * (DERIV(layer->outs[k]) +raise);
    g     = layer->grds[k];     /* first process the offset gradient */
    g[n = layer->inpcnt] -= delta;
    while (--n >= 0)            /* then process the weight gradients */
      g[n] -= layer->inps[n] * delta;
  }                             /* (but do no backpropagation) */
  return sse;                   /* return sum of squared errors */
}  /* mlp_bkprop() */

/*--------------------------------------------------------------------*/

void mlp_update (MLP *mlp)
{                               /* --- update connection weights */
  DIMID k;                      /* loop variable */

  assert(mlp);                  /* check the function argument */
  mlp->step += 1.0;             /* increment the training step */
  if (mlp->decay != 1.0)        /* if weight decay is requested */
    for (k = 0; k < mlp->wgtcnt; k++)
      mlp->wgts[k] *= mlp->decay; /* reduce all connection weights */
  updatefn[mlp->method](mlp);   /* call the weight update function */
}  /* mlp_update() */

/*--------------------------------------------------------------------*/

double mlp_sens (MLP *mlp, DIMID unit, int mode)
{                               /* --- analyze sensitivity on input */
  int      l;                   /* loop variable  for layers */
  DIMID    k, n;                /* loop variables for weights */
  MLPLAYER *layer;              /* to traverse the network layers */
  double   **w;                 /* to traverse the weight vectors */
  double   *e, *p;              /* to traverse the sensitivity values */
  double   s;                   /* resulting sensitivity */

  assert(mlp                    /* check the function arguments */
  &&    (unit >= 0) && (unit < mlp->inpcnt));
  layer = mlp->layers;          /* get the first hidden layer */
  w     = layer->wgts;          /* traverse the first hidden layer */
  e     = layer->errs;          /* and compute sensitivity values */
  for (k = layer->outcnt; --k >= 0; )
    e[k] = w[k][unit] *DERIV(layer->outs[k]);
  for (l = mlp->lyrcnt-2; --l >= 0; ) {
    p = e; ++layer;             /* traverse the remaining layers */
    e = memset(layer->errs, 0, (size_t)layer->outcnt *sizeof(double));
    w = layer->wgts;            /* clear the sensitivity values */
    for (k = layer->outcnt; --k >= 0; ) {
      for (n = layer->inpcnt; --n >= 0; )
        e[k] += w[k][n] *p[n];  /* aggregate the sensitivity values */
      e[k] *= DERIV(layer->outs[k]);  /* of the preceding layer and */
    }                           /* compute the sensitivity values */
  }                             /* of the next layer */
  p = mlp->scls;                /* traverse the output units */
  for (k = mlp->outcnt; --k >= 0; )
    e[k] = fabs(e[k] *p[k]);    /* compute absolute sensitivity value */
  k = mlp->outcnt; s = 0;       /* compute sum or maximum */
  if (mode & MLP_SUM) while (--k >= 0) { s += e[k]; }
  else                while (--k >= 0) { if (e[k] > s) s = e[k]; }
  return s;                     /* return the sensitivity */
}  /* mlp_sens() */

/*--------------------------------------------------------------------*/

int mlp_desc (MLP *mlp, FILE *file, int mode, int maxlen)
{                               /* --- describe multilayer perceptron */
  int      l;                   /* loop variable  for layers */
  DIMID    k, n;                /* loop variables for weights */
  int      len, i;              /* loop variables for comments */
  double   **wgts;              /* to traverse the weight vectors */
  char     *indent = "";        /* indentation string */

  assert(mlp && file);          /* check the function arguments */
  len = (maxlen > 0) ? maxlen -2 : 70;

  /* --- print header (as a comment) --- */
  if (mode & MLP_TITLE) {       /* if the title flag is set */
    fputs("/*", file); for (i = len; --i >= 0; ) fputc('-', file);
    fprintf(file, "\n  multilayer perceptron\n");
    for (i = len; --i >= 0; ) fputc('-', file);
    fputs("*/\n", file);        /* print a title header */
  }                             /* (as a comment) */

  fputs("mlp = {\n", file);     /* start the network description */
  indent = "  ";                /* indent the network description */

  /* --- print list of numbers of units --- */
  fprintf(file, "%sunits    = %"DIMID_FMT, indent, mlp->inpcnt);
  for (l = 0; l < mlp->lyrcnt-1; l++)
    fprintf(file, ", %"DIMID_FMT, mlp->layers[l].outcnt);
  fprintf(file, ";\n");         /* print number of units per layer */

  /* --- print list of input scalings --- */
  if (mlp->inpcnt > 0)          /* check whether there are inputs */
    nst_desc(mlp->nst, file, indent);

  /* --- print list of weight vectors --- */
  fprintf(file, "%sweights  = ", indent);
  for (l = 0; l < mlp->lyrcnt-1; l++) {   /* traverse the layers */
    if (l > 0) fprintf(file, ",\n%s           ", indent);
    fputc('{', file);           /* start a layer description */
    wgts = mlp->layers[l].wgts; /* traverse the units in a layer */
    for (k = 0; k < mlp->layers[l].outcnt; k++) {
      if (k > 0) fprintf(file, ",\n%s            ", indent);
      fputs("{ ", file);        /* print the connection weights */
      for (n = 0; n < mlp->layers[l].inpcnt; n++)
        fprintf(file, "%+.16g, ", wgts[k][n]);
      fprintf(file, "%+.16g }", wgts[k][n]);
    }                           /* print the bias value */
    fputc('}', file);           /* terminate the layer description */
  }                             /* if not last layer, start new line */
  fprintf(file, ";\n");         /* terminate the list */

  /* --- print list of output ranges --- */
  fprintf(file, "%sranges   = ", indent);
  for (k = 0; k < mlp->outcnt; k++) {
    if (k > 0) fputs(", ", file);
    if (mlp->maxs[k] < mlp->mins[k]) {
      mlp->mins[k] = ACTMIN; mlp->maxs[k] = ACTMAX; }
    fprintf(file, "[%.16g, %.16g]", mlp->mins[k], mlp->maxs[k]);
  }                             /* print the output ranges */
  fputs(";\n", file);           /* terminate the list */
  fputs("};\n", file);          /* terminate the description */

  /* --- print additional information (as a comment) --- */
  if (mode & MLP_INFO) {        /* if the add. info. flag is set */
    fputs("\n/*", file); for (i = len; --i >= 0; ) fputc('-', file);
    fprintf(file, "\n  number of inputs : %"DIMID_FMT, mlp->inpcnt);
    fprintf(file, "\n  number of outputs: %"DIMID_FMT, mlp->outcnt);
    fprintf(file, "\n  number of units  : %"DIMID_FMT, mlp->unitcnt);
    fprintf(file, "\n  number of weights: %"DIMID_FMT, mlp->wgtcnt);
    fputc('\n', file);          /* print additional network info. */
    for (i = len; --i >= 0; ) fputc('-', file);
    fputs("*/\n", file);        /* (as a comment) */
  }
  return ferror(file);          /* return write error status */
}  /* mlp_desc() */

/*--------------------------------------------------------------------*/
#ifdef MLP_PARSE

static int layers (SCANNER *scan,
                   DIMID inpcnt, DIMID outcnt, DIMID ucnts[])
{                               /* --- get number of units per layer */
  int   lyrcnt = 0;             /* number of layers */
  char  *s;                     /* end pointer for conversion */
  DIMID n;                      /* temporary buffer */

  assert(scan && ucnts);        /* check the function arguments */
  if ((scn_token(scan) != T_ID) /* check for 'units' */
  ||  (strcmp(scn_value(scan), "units") != 0))
    SCN_ERROR(scan, E_STREXP, "units");
  SCN_NEXT(scan);               /* consume 'units' */
  SCN_CHAR(scan, '=');          /* consume '=' */
  while (1) {                   /* unit counter read loop */
    if (scn_token(scan) != T_NUM)  SCN_ERROR(scan, E_NUMEXP);
    if (lyrcnt >= MLP_MAXLAYER)    SCN_ERROR(scan, E_LYRCNT);
    n = strtodim(scn_value(scan), &s);
    if (*s || (n < 0))             SCN_ERROR(scan, E_NUMBER);
    if ((lyrcnt >  0) && (n <= 0)) SCN_ERROR(scan, E_NUMBER);
    if ((lyrcnt == 0) && (inpcnt > 0) && (n != inpcnt))
      SCN_ERROR(scan, E_INPCNT);/* get and check number of inputs */
    SCN_NEXT(scan);             /* consume the number of units */
    ucnts[lyrcnt++] = n;        /* and store it */
    if (scn_token(scan) != ',') break;
    SCN_NEXT(scan);             /* if at end of list, abort the loop, */
  }                             /* otherwise consume ',' */
  if (lyrcnt < 2)                    SCN_ERROR(scan, E_LYRCNT);
  if ((outcnt > 0) && (n != outcnt)) SCN_ERROR(scan, E_OUTCNT);
  SCN_CHAR(scan, ';');          /* consume ';' */
  return lyrcnt;                /* return the number of layers */
}  /* layers() */

/*--------------------------------------------------------------------*/

static int scales (MLP *mlp, SCANNER *scan)
{                               /* --- get the input scalings */
  assert(mlp && scan);          /* check the function arguments */
  if (mlp->inpcnt < 1) return 0;/* check whether there are inputs */
  if ((scn_token(scan) != T_ID) /* check whether 'scales' follows */
  ||  (strcmp(scn_value(scan), "scales") != 0))
    mlp->nst = nst_create(mlp->inpcnt);
  else                          /* create or parse scaling parameters */
    mlp->nst = nst_parse(scan, mlp->inpcnt);
  return (mlp->nst) ? 0 : 1;    /* return an error indicator */
}  /* scales() */

/*--------------------------------------------------------------------*/

static int weights (MLP *mlp, SCANNER *scan)
{                               /* --- get the connection weights */
  int      l;                   /* loop variable  for layers */
  DIMID    k, n;                /* loop variables for weights */
  MLPLAYER *layer;              /* to traverse the network layers */
  double   **wp, *w;            /* to traverse the weights */

  assert(mlp && scan);          /* check the function arguments */
  if ((scn_token(scan) != T_ID) /* check for 'weights' */
  ||  (strcmp(scn_value(scan), "weights") != 0))
    SCN_ERROR(scan, E_STREXP, "weights");
  SCN_NEXT(scan);               /* consume 'weights' */
  SCN_CHAR(scan, '=');          /* consume '=' */
  layer = mlp->layers;          /* traverse the network layers */
  for (l = mlp->lyrcnt-1; --l >= 0; layer++) {
    SCN_CHAR(scan, '{');        /* consume '{' */
    wp = layer->wgts;           /* traverse the units of a layer */
    for (k = layer->outcnt; --k >= 0; ) {
      SCN_CHAR(scan, '{');      /* consume '{' */
      w = *wp++;                /* traverse the inputs of a unit */
      for (n = layer->inpcnt +1; --n >= 0; ) {
        if (scn_token(scan) != T_NUM) SCN_ERROR(scan, E_NUMEXP);
        *w++ = strtod(scn_value(scan), NULL);
        SCN_NEXT(scan);         /* get the connection weight */
        if (n > 0) { SCN_CHAR(scan, ','); }
      }                         /* if not bis value, consume ',' */
      SCN_CHAR(scan, '}');      /* consume '}' */
      if (k > 0) { SCN_CHAR(scan, ','); }
    }                           /* if not last unit, consume ',' */
    SCN_CHAR(scan, '}');        /* consume '}' */
    if (l > 0) { SCN_CHAR(scan, ','); }
  }                             /* if not output layer, consume ',' */
  SCN_CHAR(scan, ';');          /* consume ';' */
  return 0;                     /* return 'ok' */
}  /* weights() */

/*--------------------------------------------------------------------*/

static int ranges (MLP *mlp, SCANNER *scan)
{                               /* --- get the output ranges */
  DIMID  k;                     /* loop variable */
  double *min, *max;            /* to traverse the minima/maxima */

  assert(mlp && scan);          /* check the function arguments */
  if ((scn_token(scan) != T_ID) /* check whether 'ranges' follows */
  ||  (strcmp(scn_value(scan), "ranges") != 0))
    return 0;                   /* if not, abort the function */
  SCN_NEXT(scan);               /* consume 'ranges' */
  SCN_CHAR(scan, '=');          /* consume '=' */
  min = mlp->mins;              /* traverse the vectors of */
  max = mlp->maxs;              /* minimal and maximal output values */
  for (k = mlp->outcnt; --k >= 0; ) {
    SCN_CHAR(scan, '[');        /* consume '[' */
    if (scn_token(scan) != T_NUM) SCN_ERROR(scan, E_NUMEXP);
    *min++ = strtod(scn_value(scan), NULL);
    SCN_NEXT(scan);             /* consume the minimal output value */
    SCN_CHAR(scan, ',');        /* consume '[' */
    if (scn_token(scan) != T_NUM) SCN_ERROR(scan, E_NUMEXP);
    *max++ = strtod(scn_value(scan), NULL);
    SCN_NEXT(scan);             /* consume the maximal output value */
    SCN_CHAR(scan, ']');        /* consume '[' */
    if (k > 0) { SCN_CHAR(scan, ','); }
  }                             /* if not last unit, consume ',' */
  SCN_CHAR(scan, ';');          /* consume ';' */
  return 0;                     /* return 'ok' */
}  /* ranges() */

/*--------------------------------------------------------------------*/

static MLP* parse (SCANNER *scan, DIMID inpcnt, DIMID outcnt)
{                               /* --- parse an LVQ network */
  int   err, e;                 /* error status */
  int   lyrcnt;                 /* number of layers */
  DIMID ucnts[MLP_MAXLAYER];    /* number of units per layer */
  MLP   *mlp;                   /* parsed multilayer perceptron */

  assert(scan);                 /* check the function argument */
  scn_first(scan);              /* get first token */
  lyrcnt = layers(scan, inpcnt, outcnt, ucnts);
  if (lyrcnt < 0) return NULL;  /* get the number of units per layer */
  mlp = create(lyrcnt, ucnts);  /* create a multilayer perceptron */
  if (!mlp) return NULL;
  err      = scales(mlp, scan); /* read the scaling parameters */
  if (err && (scn_recover(scan, ';', 0, 0, 0) == T_EOF)) {
    mlp_delete(mlp); return NULL; }
  err |= e = weights(mlp,scan); /* read the connection weights */
  if (e   && (scn_recover(scan, ';', 0, 0, 0) == T_EOF)) {
    mlp_delete(mlp); return NULL; }
  err |=     ranges(mlp,scan);  /* read the output ranges */
  if (err) { mlp_delete(mlp); return NULL; }
  if (!mlp->nst) mlp->nst = nst_create(mlp->inpcnt);
  if (!mlp->nst) { mlp_delete(mlp); return NULL; }
  return mlp;                   /* return created m.layer perceptron */
}  /* parse() */

/*--------------------------------------------------------------------*/

static int header (SCANNER *scan)
{                               /* --- parse header */
  assert(scan);                 /* check the function argument */
  if ((scn_token(scan) != T_ID)
  ||  (strcmp(scn_value(scan), "mlp") != 0))
    SCN_ERROR(scan, E_STREXP, "mlp");
  SCN_NEXT(scan);               /* consume 'mlp' */
  if (scn_token(scan) == '(') { /* if a target attribute follows */
    SCN_NEXT(scan);             /* consume '(' */
    if (scn_token(scan) != T_ID) SCN_ERROR(scan, E_ATTEXP);
    SCN_NEXT(scan);             /* consume the target attribute */
    SCN_CHAR(scan, ')');        /* check for and consume ')' */
  }
  SCN_CHAR(scan, '=');          /* consume '=' */
  SCN_CHAR(scan, '{');          /* consume '{' */
  return 0;                     /* return 'ok' */
}  /* header() */

/*--------------------------------------------------------------------*/

static int trailer (SCANNER *scan)
{                               /* --- parse trailer */
  assert(scan);                 /* check the function argument */
  SCN_CHAR(scan, '}');          /* consume '}' */
  SCN_CHAR(scan, ';');          /* consume ';' */
  return 0;                     /* return 'ok' */
}  /* trailer() */

/*--------------------------------------------------------------------*/

MLP* mlp_parse (SCANNER *scan, DIMID inpcnt, DIMID outcnt, int hdr)
{                               /* --- parse a multilayer perceptron */
  MLP *mlp;                     /* parsed multilayer perceptron */

  assert(scan);                 /* check the function argument */
  scn_setmsgs(scan, msgs, (int)(sizeof(msgs)/sizeof(*msgs)));
  scn_first(scan);              /* set messages, get first token */
  if (hdr && (header(scan) != 0))
    return NULL;                /* parse the header if necessary */
  mlp = parse(scan, inpcnt, outcnt);
  if (!mlp) return NULL;        /* parse the multilayer perceptron */
  if (hdr && (trailer(scan) != 0)) {
    mlp_delete(mlp); return NULL; }
  return mlp;                   /* return the parsed network */
}  /* mlp_parse() */

#endif  /* #ifdef MLP_PARSE */
